@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 # This file is part of beets.
 # Copyright 2016, Jakob Schnitzer.
 #
@@ -13,47 +12,55 @@
 # The above copyright notice and this permission notice shall be
 # included in all copies or substantial portions of the Software.
 
-"""Update library's tags using MusicBrainz.
-"""
-from __future__ import division, absolute_import, print_function
+"""Synchronise library metadata with metadata source backends."""
 
-from beets.plugins import BeetsPlugin, apply_item_changes
-from beets import autotag, library, ui, util
-from beets.autotag import hooks
 from collections import defaultdict
 
-import re
-
-MBID_REGEX = r"(\d|\w){8}-(\d|\w){4}-(\d|\w){4}-(\d|\w){4}-(\d|\w){12}"
+from beets import autotag, library, ui, util
+from beets.autotag import hooks
+from beets.plugins import BeetsPlugin, apply_item_changes
 
 
 class MBSyncPlugin(BeetsPlugin):
     def __init__(self):
-        super(MBSyncPlugin, self).__init__()
+        super().__init__()
 
     def commands(self):
-        cmd = ui.Subcommand('mbsync',
-                            help=u'update metadata from musicbrainz')
+        cmd = ui.Subcommand("mbsync", help="update metadata from musicbrainz")
         cmd.parser.add_option(
-            u'-p', u'--pretend', action='store_true',
-            help=u'show all changes but do nothing')
+            "-p",
+            "--pretend",
+            action="store_true",
+            help="show all changes but do nothing",
+        )
         cmd.parser.add_option(
-            u'-m', u'--move', action='store_true', dest='move',
-            help=u"move files in the library directory")
+            "-m",
+            "--move",
+            action="store_true",
+            dest="move",
+            help="move files in the library directory",
+        )
         cmd.parser.add_option(
-            u'-M', u'--nomove', action='store_false', dest='move',
-            help=u"don't move files in library")
+            "-M",
+            "--nomove",
+            action="store_false",
+            dest="move",
+            help="don't move files in library",
+        )
         cmd.parser.add_option(
-            u'-W', u'--nowrite', action='store_false',
-            default=None, dest='write',
-            help=u"don't write updated metadata to files")
+            "-W",
+            "--nowrite",
+            action="store_false",
+            default=None,
+            dest="write",
+            help="don't write updated metadata to files",
+        )
         cmd.parser.add_format_option()
         cmd.func = self.func
         return [cmd]
 
     def func(self, lib, opts, args):
-        """Command handler for the mbsync function.
-        """
+        """Command handler for the mbsync function."""
         move = ui.should_move(opts.move)
         pretend = opts.pretend
         write = ui.should_write(opts.write)
@@ -66,25 +73,17 @@ class MBSyncPlugin(BeetsPlugin):
         """Retrieve and apply info from the autotagger for items matched by
         query.
         """
-        for item in lib.items(query + [u'singleton:true']):
-            item_formatted = format(item)
+        for item in lib.items(query + ["singleton:true"]):
             if not item.mb_trackid:
-                self._log.info(u'Skipping singleton with no mb_trackid: {0}',
-                               item_formatted)
+                self._log.info(
+                    "Skipping singleton with no mb_trackid: {}", item
+                )
                 continue
 
-            # Do we have a valid MusicBrainz track ID?
-            if not re.match(MBID_REGEX, item.mb_trackid):
-                self._log.info(u'Skipping singleton with invalid mb_trackid:' +
-                               ' {0}', item_formatted)
-                continue
-
-            # Get the MusicBrainz recording info.
-            track_info = hooks.track_for_mbid(item.mb_trackid)
-            if not track_info:
-                self._log.info(u'Recording ID not found: {0} for track {0}',
-                               item.mb_trackid,
-                               item_formatted)
+            if not (track_info := hooks.track_for_id(item.mb_trackid)):
+                self._log.info(
+                    "Recording ID not found: {0.mb_trackid} for track {0}", item
+                )
                 continue
 
             # Apply.
@@ -97,27 +96,15 @@ class MBSyncPlugin(BeetsPlugin):
         query and their items.
         """
         # Process matching albums.
-        for a in lib.albums(query):
-            album_formatted = format(a)
-            if not a.mb_albumid:
-                self._log.info(u'Skipping album with no mb_albumid: {0}',
-                               album_formatted)
+        for album in lib.albums(query):
+            if not album.mb_albumid:
+                self._log.info("Skipping album with no mb_albumid: {}", album)
                 continue
 
-            items = list(a.items())
-
-            # Do we have a valid MusicBrainz album ID?
-            if not re.match(MBID_REGEX, a.mb_albumid):
-                self._log.info(u'Skipping album with invalid mb_albumid: {0}',
-                               album_formatted)
-                continue
-
-            # Get the MusicBrainz album information.
-            album_info = hooks.album_for_mbid(a.mb_albumid)
-            if not album_info:
-                self._log.info(u'Release ID {0} not found for album {1}',
-                               a.mb_albumid,
-                               album_formatted)
+            if not (album_info := hooks.album_for_id(album.mb_albumid)):
+                self._log.info(
+                    "Release ID {0.mb_albumid} not found for album {0}", album
+                )
                 continue
 
             # Map release track and recording MBIDs to their information.
@@ -133,9 +120,12 @@ class MBSyncPlugin(BeetsPlugin):
             # first, if available, and recording MBIDs otherwise). This should
             # work for albums that have missing or extra tracks.
             mapping = {}
+            items = list(album.items())
             for item in items:
-                if item.mb_releasetrackid and \
-                        item.mb_releasetrackid in releasetrack_index:
+                if (
+                    item.mb_releasetrackid
+                    and item.mb_releasetrackid in releasetrack_index
+                ):
                     mapping[item] = releasetrack_index[item.mb_releasetrackid]
                 else:
                     candidates = track_index[item.mb_trackid]
@@ -145,17 +135,19 @@ class MBSyncPlugin(BeetsPlugin):
                         # If there are multiple copies of a recording, they are
                         # disambiguated using their disc and track number.
                         for c in candidates:
-                            if (c.medium_index == item.track and
-                                    c.medium == item.disc):
+                            if (
+                                c.medium_index == item.track
+                                and c.medium == item.disc
+                            ):
                                 mapping[item] = c
                                 break
 
             # Apply.
-            self._log.debug(u'applying changes to {}', album_formatted)
+            self._log.debug("applying changes to {}", album)
             with lib.transaction():
                 autotag.apply_metadata(album_info, mapping)
                 changed = False
-                # Find any changed item to apply MusicBrainz changes to album.
+                # Find any changed item to apply changes to album.
                 any_changed_item = items[0]
                 for item in items:
                     item_changed = ui.show_model_changes(item)
@@ -171,10 +163,10 @@ class MBSyncPlugin(BeetsPlugin):
                 if not pretend:
                     # Update album structure to reflect an item in it.
                     for key in library.Album.item_keys:
-                        a[key] = any_changed_item[key]
-                    a.store()
+                        album[key] = any_changed_item[key]
+                    album.store()
 
                     # Move album art (and any inconsistent items).
                     if move and lib.directory in util.ancestry(items[0].path):
-                        self._log.debug(u'moving album {0}', album_formatted)
-                        a.move()
+                        self._log.debug("moving album {}", album)
+                        album.move()
